@@ -1,64 +1,30 @@
 """This module contains the InstaClient class"""
-import json
-from random import randrange
 from selenium import webdriver
+from selenium.common.exceptions import WebDriverException
 from selenium.webdriver.remote.webelement import WebElement
 from selenium.webdriver.support.wait import WebDriverWait
 from selenium.webdriver.common.by import By
 from selenium.webdriver.common.keys import Keys
 from selenium.webdriver.support import expected_conditions as EC
 from selenium.common.exceptions import ElementClickInterceptedException, NoSuchElementException, TimeoutException        
-import time, os
-
-from urllib3 import request
+import time, os, requests, concurrent.futures
+from random import randrange
 
 from instaclient.utilities.utilities import *
 from instaclient.errors import *
 from instaclient.client.paths import Paths
 from instaclient.client.urls import ClientUrls, GraphUrls
-from instaclient.client.scraper import Scraper
+from instaclient.client.notiscraper import NotificationScraper
+from instaclient.client.tagscraper import TagScraper
 
-class InstaClient(Scraper):
+
+class InstaClient(NotificationScraper, TagScraper):
     CHROMEDRIVER=1
     LOCAHOST=1
     WEB_SERVER=2
     PIXEL_SCROLL=3
     END_PAGE_SCROLL=4
     PAGE_DOWN_SCROLL=5
-    # INSTACLIENT DECORATOR
-    def __manage_driver(func):
-        @wraps(func)
-        def wrapper(self, *args, **kwargs):
-            time.sleep(random.randint(1, 2))
-            self.logger.debug('INSTACLIENT: Mangage Driver, func: {}'.format(func.__name__))
-            if not self.driver:
-                login = True
-                if func.__name__ in ('login', 'resend_verification_code', 'input_security_code', 'input_verification_code', 'logout'):
-                    login = False
-                self.__init_driver(login, func=func.__name__)
-
-            error = False
-            result = None
-            try:
-                result = func(self, *args, **kwargs)
-                time.sleep(1)
-            except Exception as exception:
-                error = exception
-            
-            discard = kwargs.get('discard_driver')
-            if discard is not None:
-                if discard:
-                    self.__discard_driver()
-            elif len(args) > 0 and isinstance(args[-1], bool):
-                if args[-1]:
-                    self.__discard_driver()
-            
-            time.sleep(random.randint(1, 2))
-            if error:
-                raise error
-            else:
-                return result
-        return wrapper
     
     # INIT
     def __init__(self, driver_type: int=CHROMEDRIVER, host_type:int=LOCAHOST, driver_path=None, init_driver=False, debug=False, error_callback=None, localhost_headless=False):
@@ -92,19 +58,65 @@ class InstaClient(Scraper):
                 raise InvalidErrorCallbackError()
         self.error_callback = error_callback
         self.localhost_headless = localhost_headless
-        self.logger = logging.getLogger(__name__)
-        if debug:
-            self.logger.setLevel(logging.DEBUG)
         self.logged_in = False
         self.driver = None
         self.username = None
         self.password = None
+        self.threads = []
+
+        self.logger = logging.getLogger(__name__)
+        if debug:
+            self.logger.setLevel(logging.DEBUG)
+        else:
+            self.logger.setLevel(logging.INFO)
+
+        NotificationScraper.__init__(self, self.logger)
+        TagScraper.__init__(self, self.logger)
+
         if init_driver:
             self.__init_driver(func='__init__')
 
+
+    # INSTACLIENT DECORATOR
+    def __manage_driver(init_driver=True, login=True):
+        def outer(func):
+            @wraps(func)
+            def wrapper(self, *args, **kwargs):
+                time.sleep(random.randint(1, 2))
+                self.logger.debug('INSTACLIENT: Mangage Driver, func: {}'.format(func.__name__))
+                if init_driver:
+                    if not self.driver:
+                        if login and (self.username is None or self.password is None):
+                            raise NotLoggedInError()
+                        self.__init_driver(login, func=func.__name__)
+
+                error = False
+                result = None
+                try:
+                    result = func(self, *args, **kwargs)
+                    time.sleep(1)
+                except Exception as exception:
+                    error = exception
+                
+                discard = kwargs.get('discard_driver')
+                if discard is not None:
+                    if discard:
+                        self.__discard_driver()
+                elif len(args) > 0 and isinstance(args[-1], bool):
+                    if args[-1]:
+                        self.__discard_driver()
+                
+                time.sleep(random.randint(1, 2))
+                if error:
+                    raise error
+                else:
+                    return result
+            return wrapper
+        return outer
+    
     # INSTAGRAM FUNCTIONS
     # LOGIN PROCEDURE
-    @__manage_driver
+    @__manage_driver(login=False)
     def check_status(self, discard_driver:bool=False):
         """
         Check if account is currently logged in. Returns True if account is logged in. Sets the `instaclient.logged_in` variable accordingly.
@@ -136,7 +148,7 @@ class InstaClient(Scraper):
         return result
 
 
-    @__manage_driver
+    @__manage_driver(login=False)
     def login(self, username:str, password:str, check_user:bool=True, discard_driver:bool=False):
         """
         Sign Into Instagram with credentials. Go through 2FA if necessary. Sets the InstaClient variable `InstaClient.logged_in` to True if login was successful.
@@ -258,7 +270,7 @@ class InstaClient(Scraper):
         return self.logged_in
 
 
-    @__manage_driver
+    @__manage_driver(login=False)
     def resend_security_code(self):
         """
         Resend security code if code hasn't been sent successfully. The code is used to verify the login attempt if `instaclient.errors.common.SuspiciousLoginAttemptError` is raised.
@@ -292,7 +304,7 @@ class InstaClient(Scraper):
             return False
 
 
-    @__manage_driver
+    @__manage_driver(login=False)
     def input_security_code(self, code:int or str, discard_driver:bool=False):
         """
         Complete login procedure started with `InstaClient.login()` and insert security code required if `instaclient.errors.common.SuspiciousLoginAttemptError` is raised. Sets `InstaClient.logged_in` attribute to True if login was successful.
@@ -330,7 +342,7 @@ class InstaClient(Scraper):
         return self.logged_in
 
 
-    @__manage_driver
+    @__manage_driver(login=False)
     def input_verification_code(self, code:int or str, discard_driver:bool=False):
         """
         Complete login procedure started with `InstaClient.login()` and insert 2FA security code. Sets `instaclient.logged_in` to True if login was successful.
@@ -363,7 +375,7 @@ class InstaClient(Scraper):
             return self.logged_in
 
 
-    @__manage_driver
+    @__manage_driver(login=False)
     def logout(self, discard_driver:bool=False):
         """
         Check if the client is currently connected to Instagram and logs of the current InstaClient session.
@@ -392,7 +404,7 @@ class InstaClient(Scraper):
             return True
 
 
-    @__manage_driver
+    @__manage_driver(login=False)
     def is_valid_user(self, user:str, nav_to_user:bool=True, discard_driver:bool=False):
         """
         is_valid_user Checks if a given username is a valid Instagram user.
@@ -446,7 +458,7 @@ class InstaClient(Scraper):
 
 
     # FOLLOW PROCEDURE
-    @__manage_driver
+    @__manage_driver()
     def follow_user(self, user:str, nav_to_user:bool=True, discard_driver:bool=False):
         """
         follow_user follows the instagram user that matches the username in the `user` attribute.
@@ -488,7 +500,7 @@ class InstaClient(Scraper):
             raise FollowRequestSentError(user)
 
     
-    @__manage_driver
+    @__manage_driver()
     def unfollow_user(self, user:str, nav_to_user=True, check_user=True, discard_driver:bool=False):
         """
         Unfollows user(s)
@@ -512,7 +524,7 @@ class InstaClient(Scraper):
 
 
     # USER DATA PRODECURES
-    @__manage_driver
+    @__manage_driver()
     def get_user_images(self, user:str, discard_driver:bool=False):
         """
         Get all images from a users profile.
@@ -538,24 +550,74 @@ class InstaClient(Scraper):
 
         img_srcs = list(set(img_srcs)) # clean up duplicates
         return img_srcs
-    
 
-    @__manage_driver
-    def scrape_followers(self, user:str, check_user=True, discard_driver:bool=False):
+    
+    @__manage_driver(login=False)
+    def scrape_followers(self, user:str, check_user=True, max_wait_time:int=250, callback_frequency:int=20, callback=None, discard_driver=False):
         """
         scrape_followers Scrape an instagram user's followers and return them as a list of strings.
 
         Args:
             user (str): User to scrape
             check_user (bool, optional): If set to True, checks if the `user` is a valid instagram username. Defaults to True.
-            discard_driver (bool, optional): If set to True, the `driver` will be discarded at the end of the method. Defaults to False.
+            max_wait_time (int, optional): Time to wait (in seconds) while loading before stopping the method and returning the results. Defaults to 300 (seconds)
+            callback_frequency (int, optional): Time (in seconds) between updates
+            callback (function): Function with no parameters that gets called with the frequency set by ``callback_frequency``
 
         Returns:
             list: List of instagram usernames
         """
+        if callback and not callable(callback):
+            raise InvalidErrorCallbackError()
+
+        with concurrent.futures.ThreadPoolExecutor(max_workers=2, thread_name_prefix='scrape') as executor:
+            scrape_future = executor.submit(self.__scrape_followers, user, check_user, max_wait_time)
+            timer_future = executor.submit(self.__scrape_timer, max_wait_time, callback_frequency, callback)
+            result = scrape_future.result()
+            return result
+
+    
+    def __scrape_timer(self, max_wait_time:int, callback_frequency:int, callback=None):
+        tic = time.time()
+        final_time = tic + max_wait_time
+        self.logger.info('Initiating scrape...'.format(callback_frequency))
+        while True:
+            
+            self.logger.info('Scraping followers... waited {} seconds'.format(callback_frequency))
+            if callback:
+                callback()
+            toc = time.time()
+            self.logger.debug('TIMER: {} | Difference: {} | Total Difference: {}'.format(toc, (toc - tic), (final_time - toc)))
+            if toc > final_time:
+                break
+            elif (final_time - toc) < 0:
+                time.sleep(final_time - toc)
+            else:
+                tic = toc
+                time.sleep(callback_frequency)
+        self.logger.debug('Finished Timer')
+        return True
+
+        
+    def __scrape_followers(self, user:str, check_user=True, max_waiting_time:int=250):
+        """
+        __scrape_followers Scrape an instagram user's followers and return them as a list of strings.
+
+        Args:
+            user (str): User to scrape
+            check_user (bool, optional): If set to True, checks if the `user` is a valid instagram username. Defaults to True.
+
+        Returns:
+            list: List of instagram usernames
+        """
+        # Set starting time:
+        tic = time.time()
+        final_time = tic + max_waiting_time
         # Nav to user page
         self.nav_user(user, check_user=check_user)
+        self.logger.debug('Navigated to User Page')
         # Find Followers button/link
+        self.logger.debug('Found Followers Button')
         followers_btn:WebElement = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.FOLLOWERS_BTN)), url=ClientUrls.NAV_USER.format(user))
         # Start scraping
         followers = []
@@ -565,24 +627,40 @@ class InstaClient(Scraper):
         # Load all followers
         followers = []
         main:WebElement = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.FOLLOWERS_LIST_MAIN)))
+        self.scroll(self.END_PAGE_SCROLL, times=10, interval=0)
+        main:WebElement = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.FOLLOWERS_LIST_MAIN)))
         size = main.size.get('height')
-        time.sleep(15)
         while True:
+            self.logger.debug('INSTACLIENT: Loop')
             main:WebElement = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.FOLLOWERS_LIST_MAIN)), wait_time=3)
             new_size = main.size.get('height')
-            if new_size > 60000:
-                break
             if new_size > size:
                 size = new_size
-                time.sleep(15)
+                toc = time.time()
+                self.logger.debug('SCRAPER: {} | Difference: {} | Total Difference: {}'.format(toc, (toc - tic), (final_time - toc)))
+                if toc > final_time:
+                    self.logger.debug('BREAKING LOOP')
+                    break
+                tic = toc
+                self.scroll(self.END_PAGE_SCROLL, times=10, interval=0)
                 continue
             else:
-                break
+                self.scroll(self.PAGE_DOWN_SCROLL)
+                main:WebElement = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.FOLLOWERS_LIST_MAIN)), wait_time=3)
+                new_size = main.size.get('height')
+                if new_size > size:
+                    continue
+                else:
+                    self.logger.debug('BREAKING LOOP')
+                    break
+        self.logger.debug('Finished Scrolling. Loading followers div...')
         try:
             followers_list:WebElement = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.FOLLOWERS_LIST)), wait_time=3)
         except NoSuchElementException:
-            return self.scrape_followers(user, check_user, discard_driver)
+            self.logger.warn('Followers List not found... Retrying method')
+            return self.scrape_followers(user, check_user)
 
+        self.logger.info('Followers loaded. Saving and returing')
         divs = followers_list.find_elements_by_xpath(Paths.FOLLOWER_USER_DIV)
         for div in divs:
             try:
@@ -595,7 +673,7 @@ class InstaClient(Scraper):
 
 
     # ENGAGEMENT PROCEDURES
-    @__manage_driver
+    @__manage_driver()
     def scroll(self, mode=PAGE_DOWN_SCROLL, size=500, times=1, interval=3):
         """
         Scrolls to the bottom of a users page to load all of their media
@@ -622,20 +700,22 @@ class InstaClient(Scraper):
         return False
 
 
-    @__manage_driver
+    @__manage_driver()
     def like_feed_posts(self, count):
         self.logger.debug('INSTACLIENT: like_feed_posts')
 
     
-    @__manage_driver
-    def check_notifications(self):
+    @__manage_driver()
+    def check_notifications(self, types:list=None, count:int=None):
         self.logger.debug('INSTACLIENT: check_notifications')
-        self.driver.get(GraphUrls.NOTIFICATIONS_GURL)
-        source = request
-        notifications = self.__iterate_nodes()
+        self.driver.get(GraphUrls.GRAPH_ACTIVITY)
+        element:WebElement = self.__find_element(EC.presence_of_element_located((By.XPATH, Paths.QUERY_ELEMENT)))
+        source = element.text        
+        notifications = self._scrape_notifications(source, viewer=self.username, types=types, count=count)
+        return notifications
 
 
-    @__manage_driver
+    @__manage_driver()
     def like_latest_posts(self, user:str, n_posts:int, like:bool=True, discard_driver:bool=False):
         """
         Likes a number of a users latest posts, specified by n_posts.
@@ -668,7 +748,7 @@ class InstaClient(Scraper):
             self.driver.find_elements_by_class_name('ckWGn')[0].click()
 
 
-    @__manage_driver
+    @__manage_driver()
     def send_dm(self, user:str, message:str, discard_driver:bool=False):
         """
         Send an Instagram Direct Message to a user. if `check_user` is set to True, the `user` argument will be checked to validate whether it is a real instagram username.
@@ -693,7 +773,14 @@ class InstaClient(Scraper):
             raise error
 
 
-    #@__manage_driver
+    @__manage_driver()
+    def get_hashtag(self, tag, count):
+        self.logger.debug('INSTACLIENT: check_notifications')
+        tag = self._scrape_tag(tag, None)
+        return tag
+    
+    
+    #@__manage_driver()
     #def comment_post(self, text):
         #"""
         #Comments on a post that is in modal form
@@ -707,7 +794,7 @@ class InstaClient(Scraper):
         #self.logger.debug('Commentd.')
 
 
-    """ @__manage_driver # TODO
+    """ @__manage_driver() # TODO
     def scrape_dms(self, discard_driver:bool=False):
         if not self.driver:
             self.__init_driver()
@@ -742,8 +829,8 @@ class InstaClient(Scraper):
 
                 
     # NAVIGATION PROCEDURES
-    @__manage_driver
-    def search_tag(self, tag:str, discard_driver:bool=False):
+    @__manage_driver()
+    def nav_tag(self, tag:str, discard_driver:bool=False):
         """
         Naviagtes to a search for posts with a specific tag on IG.
 
@@ -761,7 +848,7 @@ class InstaClient(Scraper):
             return True
 
 
-    @__manage_driver
+    @__manage_driver()
     def nav_user(self, user:str, check_user:bool=True):
         """
         Navigates to a users profile page
@@ -781,7 +868,7 @@ class InstaClient(Scraper):
             return self.is_valid_user(user=user, nav_to_user=False)
         
 
-    @__manage_driver
+    @__manage_driver()
     def nav_user_dm(self, user:str, check_user:bool=True):
         """
         Open DM page with a specific user
@@ -992,6 +1079,8 @@ class InstaClient(Scraper):
                     chrome_options.add_argument("--headless")
                     chrome_options.add_argument("--disable-dev-shm-usage")
                     chrome_options.add_argument("--no-sandbox")
+                    chrome_options.add_argument("--disable-setuid-sandbox") 
+                    chrome_options.add_argument("--remote-debugging-port=9222")
                     chrome_options.add_experimental_option("excludeSwitches", ["enable-automation"])
                     chrome_options.add_experimental_option('useAutomationExtension', False)
                     self.driver = webdriver.Chrome(executable_path=os.environ.get("CHROMEDRIVER_PATH"), chrome_options=chrome_options)
